#include "vulkan_allocator.hpp"

#include "exceptions.hpp"
#include "vulkan_instance.hpp"

#include <glm/vector_relational.hpp>
#include <gz-util/util/string_conversion.hpp>

namespace gz::vlk {
	std::string MemoryBlock::toString() const {
		return std::string("<offset: " + gz::toHexString(offset, 0) + ", size: " + gz::toHexString(size, 0) + ", free: " + gz::toString(free) + ">");
	}


	DeviceMemory::DeviceMemory(vk::DeviceSize size_) {
		size = size_;
		memory = VK_NULL_HANDLE;
		blocks.emplace_back(MemoryBlock{size, 0, true});
	};


	VulkanAllocator::VulkanAllocator(VulkanInstance& instance)
		: vk(instance)
	{
        LogCreateInfo logCI {
        	.logfile = "vulkan_allocator.log",
        	.storeLog = false,
        	.prefix = "Allocator",
        	.prefixColor = Color::BLACK,
        	.timeColor = settings::VULKAN_MESSAGE_TIME_COLOR,
		};
        aLog = Log(std::move(logCI));

	}

	void VulkanAllocator::allocate(const vk::MemoryAllocateInfo& allocI, const vk::MemoryRequirements2& memReq, MemoryInfo& memoryInfo) {
		aLog.log0("allocate: Requesting memory with ( size", toHexString(allocI.allocationSize), 
				"), ( memoryTypeIndex", allocI.memoryTypeIndex, 
				"), ( alignment", toHexString(memReq.memoryRequirements.alignment), ")");
		// go through allocated memories with matching memoryType
		for (auto deviceMemory = memory[allocI.memoryTypeIndex].begin(); deviceMemory != memory[allocI.memoryTypeIndex].end(); deviceMemory++) {
			/* auto bestBlockIt = deviceMemory.blocks.end(); */
			// go through blocks in memory
			for (auto block = deviceMemory->blocks.begin(); block != deviceMemory->blocks.end(); block++) {
				if (!block->free) { continue; }
				if (block->size >= allocI.allocationSize) {
					// check if alignment is ok
					if (block->offset % memReq.memoryRequirements.alignment != 0) {
						// move non-aligned space to previous block
						// new offset === offset + (align - offset % align)
						size_t moveToPreviousBlock = memReq.memoryRequirements.alignment - block->offset % memReq.memoryRequirements.alignment;
						// check if still large enough
						if (block->size - moveToPreviousBlock < allocI.allocationSize) { continue; }
						block--;
						block->size += moveToPreviousBlock;
						block++;
						block->offset += moveToPreviousBlock;
						block->size -= moveToPreviousBlock;
					}
					// if the block is larger than the needed size, split the block
					if (block->size > allocI.allocationSize) {
						// emplace free part of block after used part of block
						auto newBlock = ++block;
						block--;
						deviceMemory->blocks.emplace(newBlock,
								MemoryBlock{ block->size - allocI.allocationSize, block->offset + allocI.allocationSize, true });
						/* block--; */
						block->size = allocI.allocationSize;
					}
					block->free = false;
					memoryInfo.memory = deviceMemory->memory;
					memoryInfo.offset = block->offset;
					memoryInfo.memoryTypeIndex = allocI.memoryTypeIndex;

					/* aLog.log0("allocate: (memory", memoryInfo.memoryTypeIndex, "-", gz::toString(deviceMemory - memory[memoryInfo.memoryTypeIndex].begin()) + ") Blocks:", deviceMemory->blocks); */
					return;
				}
			}			
		}

		// if nothing is found, allocate new memory
		memory[allocI.memoryTypeIndex].emplace_back(DeviceMemory(allocI.allocationSize * TODO_ALLOCATION_SIZE_MULTIPLIIER));
		vk::MemoryAllocateInfo allocI_{};
		allocI_.memoryTypeIndex = allocI.memoryTypeIndex;
		allocI_.allocationSize = memory[allocI.memoryTypeIndex].back().size;
		aLog.log0("allocate: Allocating new memory of size:", gz::toHexString(allocI_.allocationSize, 0), "(memory", memoryInfo.memoryTypeIndex, "-", gz::toString(memory[memoryInfo.memoryTypeIndex].size()) + ")");
		vk::Result result = vk.getDevice().allocateMemory(&allocI_, NO_ALLOC, &memory[allocI.memoryTypeIndex].back().memory);
		if (result != vk::Result::eSuccess) {
			throw getVkException(result, "Failed to allocate memory", "VulkanAllocator::allocate");
		}

		// return block of new memory
		DeviceMemory& deviceMemory = memory[allocI.memoryTypeIndex].back();
		deviceMemory.blocks.front().offset = allocI.allocationSize;
		deviceMemory.blocks.front().size -= allocI.allocationSize;
		deviceMemory.blocks.emplace_front(MemoryBlock{ allocI.allocationSize, 0, false });

		// alignment always satisfied with vk::allocateMemory() and offset 0
		memoryInfo.memory = deviceMemory.memory;
		memoryInfo.offset = 0;
		memoryInfo.memoryTypeIndex = allocI.memoryTypeIndex;
		/* aLog.log0("allocate: (memory", memoryInfo.memoryTypeIndex, "-", gz::toString(memory[memoryInfo.memoryTypeIndex].size()) + ") Blocks:", deviceMemory.blocks); */
	}


	void VulkanAllocator::free(MemoryInfo& memoryInfo) {
		// go through allocated memories with matching memoryType
		for (auto deviceMemory = memory[memoryInfo.memoryTypeIndex].begin(); deviceMemory != memory[memoryInfo.memoryTypeIndex].end(); deviceMemory++) {
			if (deviceMemory->memory != memoryInfo.memory) { continue; }
			// go through blocks in memory
			for (auto block = deviceMemory->blocks.begin(); block != deviceMemory->blocks.end(); block++) {
				if (block->offset != memoryInfo.offset) { continue; }
				aLog.log0("free: (memory", memoryInfo.memoryTypeIndex, "-", gz::toString(deviceMemory - memory[memoryInfo.memoryTypeIndex].begin()) + ") Freeing block at offset:", gz::toHexString(memoryInfo.offset, 0));
				block->free = true;
				if (block != deviceMemory->blocks.begin()) {
					// merge with previous block, if free
					auto otherBlock = block;
					otherBlock--;
					if (otherBlock->free) {
						otherBlock->size += block->size;
						deviceMemory->blocks.erase(block);
						block = otherBlock;
					}
				}
				if (block != --(deviceMemory->blocks.end())) {
					// merge with next block, if free
					auto otherBlock = block;
					otherBlock++;
					if (otherBlock->free) {
						block->size += otherBlock->size;
						deviceMemory->blocks.erase(otherBlock);
					}
				}

				memoryInfo.memory = VK_NULL_HANDLE;
				memoryInfo.offset = 0; 
				
				/* aLog.log0("free: (memory", memoryInfo.memoryTypeIndex, "-", gz::toString(deviceMemory - memory[memoryInfo.memoryTypeIndex].begin()) + ") Blocks:", deviceMemory->blocks); */
				// if now there is only one free block, everything is free -> deallocate memory
				if (deviceMemory->blocks.size() == 1) {
					aLog.log0("free: Deallocating (memory", memoryInfo.memoryTypeIndex, "-", gz::toString(deviceMemory - memory[memoryInfo.memoryTypeIndex].begin()) + ") of size:", gz::toHexString(deviceMemory->size, 0));
					vk.getDevice().freeMemory(deviceMemory->memory, NO_ALLOC);
					memory[memoryInfo.memoryTypeIndex].erase(deviceMemory);
				}
				return;
			}
		}
		throw VkUserError("Trying to free block that was not allocated from this VulkanAllocator", "VulkanAllocator::free");
	}


	void VulkanAllocator::cleanup() {
		aLog.log0("cleanup: cleaning up");
		for (auto memType = memory.begin(); memType != memory.end(); memType++) {
			for (auto deviceMemory = memType->second.begin(); deviceMemory != memType->second.end(); deviceMemory++) {
				// check if all blocks are free
				for (auto block = deviceMemory->blocks.begin(); block != deviceMemory->blocks.end(); block++) {
					if (block->free) { continue; }
					aLog.warning("cleanup: (memory", memType->first, "-", gz::toString(deviceMemory - memType->second.begin()) + ") Block not freed: ", *block);
				}
				aLog.log0("free: Deallocating (memory", memType->first, "-", gz::toString(deviceMemory - memType->second.begin()) + ") of size:", gz::toHexString(deviceMemory->size, 0));
				vk.getDevice().freeMemory(deviceMemory->memory, NO_ALLOC);
			}
		}
		memory.clear();
	}


}  // namespace gz::vlk
